/**************************************************************
SVPFlow1 - motion vectors search for SmoothVideo Project (SVP)
Copyright (c)2016 Chainik, SVP-Team - https://www.svp-team.com

Unified interface between core logic and native Avisynth,
Vapoursynth, etc interfaces.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
http://www.gnu.org/copyleft/gpl.html
*****************************************************************/

#include "basefilter.h"

#ifdef AVISYNTH_PLUGIN
#	define YRPLAN(a) (a)->GetReadPtr(PLANAR_Y)
#	define YWPLAN(a) (a)->GetWritePtr(PLANAR_Y)
#	define URPLAN(a) (a)->GetReadPtr(PLANAR_U)
#	define UWPLAN(a) (a)->GetWritePtr(PLANAR_U)
#	define VRPLAN(a) (a)->GetReadPtr(PLANAR_V)
#	define VWPLAN(a) (a)->GetWritePtr(PLANAR_V)
#	define YPITCH(a) (a)->GetPitch(PLANAR_Y)
#	define UPITCH(a) (a)->GetPitch(PLANAR_U)
#	define VPITCH(a) (a)->GetPitch(PLANAR_V)
#endif

#ifdef AVISYNTH_PLUGIN
BaseFilter::Frame::Frame(PVideoFrame frm, int bpp)
{
    avsFrame=frm;
    this->bpp = bpp;
}
#endif

#ifdef VAPOURSYNTH_PLUGIN
BaseFilter::Frame::Frame(VSFrameRef *frm,const VSAPI *vsapi)
{
    this->vsapi=vsapi;
    vsFrame=frm;
    writeable=true;
    releaseOnDelete = false;
	durNum = 0; durDen = 1;
}

BaseFilter::Frame::Frame(const VSFrameRef *frm,const VSAPI *vsapi)
{
    this->vsapi=vsapi;
    vsFrame=const_cast<VSFrameRef*>(frm);
    writeable=false;
    releaseOnDelete = true;
	durNum = 0; durDen = 1;
}

#endif

BaseFilter::Frame::~Frame()
{
#ifdef AVISYNTH_PLUGIN
    avsFrame = 0;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    if(releaseOnDelete)
        vsapi->freeFrame(vsFrame);
    vsFrame=0;
#endif
}

bool BaseFilter::Frame::isNull() const
{
#ifdef AVISYNTH_PLUGIN
    return avsFrame == 0;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    return vsFrame == 0;
#endif
}

void BaseFilter::Frame::setReleaseOnDelete(bool on)
{
#ifdef VAPOURSYNTH_PLUGIN
    releaseOnDelete = on;
#endif
}

#ifdef VAPOURSYNTH_PLUGIN
void BaseFilter::Frame::getDuration(int64_t& num, int64_t& den) const
{
	num = durNum; den = durDen;
}

void BaseFilter::Frame::setDuration(int64_t num, int64_t den)
{
	durNum = num; durDen = den;
    if(durNum<0 || durDen<0)
        { durNum = 0; durDen = 1; }
}

int64_t BaseFilter::Frame::getPTS() const
{ return pts; }

void BaseFilter::Frame::setPTS(int64_t _pts)
{ pts = _pts; }
#endif

int BaseFilter::Frame::width() const
{
#ifdef AVISYNTH_PLUGIN
    return isNull() ? 0:avsFrame->GetRowSize()/bpp;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    return isNull() ? 0:vsapi->getFrameWidth(vsFrame,0);
#endif
}

int BaseFilter::Frame::height() const
{
#ifdef AVISYNTH_PLUGIN
    return isNull() ? 0:avsFrame->GetHeight();
#endif
#ifdef VAPOURSYNTH_PLUGIN
    return isNull() ? 0:vsapi->getFrameHeight(vsFrame,0);
#endif
}

int BaseFilter::Frame::pitchY() const
{
#ifdef AVISYNTH_PLUGIN
    return isNull() ? 0:avsFrame->GetPitch(PLANAR_Y);
#endif
#ifdef VAPOURSYNTH_PLUGIN
    return isNull() ? 0:vsapi->getStride(vsFrame,0);
#endif
}

int BaseFilter::Frame::pitchUV() const
{
#ifdef AVISYNTH_PLUGIN
    return isNull() ? 0:avsFrame->GetPitch(PLANAR_U);
#endif
#ifdef VAPOURSYNTH_PLUGIN
    return isNull() ? 0:vsapi->getStride(vsFrame,1);
#endif
}

bool BaseFilter::Frame::read(uint8_t** planeY,uint8_t** planeU,uint8_t** planeV) const
{
#ifdef AVISYNTH_PLUGIN
    if(!avsFrame) return false;
    if(planeY) *planeY = const_cast<uint8_t*>(YRPLAN(avsFrame));
    if(planeU) *planeU = const_cast<uint8_t*>(URPLAN(avsFrame));
    if(planeV) *planeV = const_cast<uint8_t*>(VRPLAN(avsFrame));
#endif
#ifdef VAPOURSYNTH_PLUGIN
    if(!vsFrame) return false;
    if(planeY) *planeY = const_cast<uint8_t*>(vsapi->getReadPtr(vsFrame,0));
    if(planeU) *planeU = const_cast<uint8_t*>(vsapi->getReadPtr(vsFrame,1));
    if(planeV) *planeV = const_cast<uint8_t*>(vsapi->getReadPtr(vsFrame,2));
#endif
    return true;
}

bool BaseFilter::Frame::write(uint8_t** planeY,uint8_t** planeU,uint8_t** planeV)
{
#ifdef AVISYNTH_PLUGIN
    if(!avsFrame || !avsFrame->IsWritable()) return false;
    if(planeY) *planeY = YWPLAN(avsFrame);
    if(planeU) *planeU = UWPLAN(avsFrame);
    if(planeV) *planeV = VWPLAN(avsFrame);
#endif
#ifdef VAPOURSYNTH_PLUGIN
    if(!vsFrame || !writeable) return false;
    if(planeY) *planeY = vsapi->getWritePtr(vsFrame,0);
    if(planeU) *planeU = vsapi->getWritePtr(vsFrame,1);
    if(planeV) *planeV = vsapi->getWritePtr(vsFrame,2);
#endif
    return true;
}

//////////////////////////////

#ifdef AVISYNTH_PLUGIN
BaseFilter::Video::Video(int w, int h, PixelType type)
{
    setSize(w,h);
    setPixelType(type);

    avs_vi.num_audio_samples=0;
    avs_vi.nchannels=0;
}

BaseFilter::Video::Video(const VideoInfo& vi)
{ avs_vi = vi; }
#endif

#ifdef VAPOURSYNTH_PLUGIN
BaseFilter::Video::Video(int w,int h,PixelType type, const VSAPI *vsapi, VSCore *vscore)
{
    this->vsapi=vsapi;
    this->vscore=vscore;

    setSize(w,h);
    setPixelType(type);
}

BaseFilter::Video::Video(const VSVideoInfo& vi, const VSAPI *vsapi, VSCore *vscore)
{
    vs_vi = vi;
    this->vsapi=vsapi;
    this->vscore=vscore;
}
#endif

int BaseFilter::Video::width() const
{
#ifdef AVISYNTH_PLUGIN
    return  pixelType()==PixelData ? avs_vi.width*4 : avs_vi.width;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    return vs_vi.width;
#endif
}

int BaseFilter::Video::height() const
{
#ifdef AVISYNTH_PLUGIN
    return avs_vi.height;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    return vs_vi.height;
#endif
}

void BaseFilter::Video::setSize(int w,int h)
{
#ifdef AVISYNTH_PLUGIN
    bool data = pixelType()==PixelData;
    avs_vi.width = data ? w/4 : w;
    avs_vi.height = h;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    vs_vi.width = w;
    vs_vi.height = h;
#endif
}

void BaseFilter::Video::frameRate(uint32_t& num,uint32_t& den) const
{
#ifdef AVISYNTH_PLUGIN
    num = avs_vi.fps_numerator;
    den = avs_vi.fps_denominator;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    num = vs_vi.fpsNum;
    den = vs_vi.fpsDen;
#endif
}

void BaseFilter::Video::setFrameRate(uint32_t num,uint32_t den)
{
#ifdef AVISYNTH_PLUGIN
    avs_vi.SetFPS(num,den);
#endif
#ifdef VAPOURSYNTH_PLUGIN
    //taken from avisynth.h - VideoInfo::SetFPS()
    if(!num || !den)
    {
        vs_vi.fpsNum=0;
        vs_vi.fpsDen=1;
    }
    else
    {
        uint32_t x=num, y=den;
        while(y) {   // find gcd
            uint32_t t = x%y; x = y; y = t;
        }
        vs_vi.fpsNum = num/x;
        vs_vi.fpsDen = den/x;
    }
#endif
}

BaseFilter::Video::PixelType BaseFilter::Video::pixelType() const
{
#ifdef AVISYNTH_PLUGIN
    if(avs_vi.IsYV12()) return PixelVideo_YV12;
    else if(avs_vi.IsYUV())
    {
        if(avs_vi.BitsPerComponent()==10) return PixelVideo_YUV_P10;
        else if(avs_vi.BitsPerComponent()==16) return PixelVideo_YUV_P16;
    }
    else if(avs_vi.IsRGB32()) return PixelData;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    switch(vs_vi.format->id)
    {
    case pfYUV420P8: return PixelVideo_YV12;
    case pfYUV420P10: return PixelVideo_YUV_P10;
    case pfYUV420P16: return PixelVideo_YUV_P16;
    case pfGray8: return PixelData;
    default:;
    }
#endif
    return PixelUnknown;
}

void BaseFilter::Video::setPixelType(BaseFilter::Video::PixelType type)
{
#ifdef AVISYNTH_PLUGIN
    switch(type)
    {
    case PixelVideo_YV12: avs_vi.pixel_type=VideoInfo::CS_YV12; break;
    case PixelVideo_YUV_P10: avs_vi.pixel_type=VideoInfo::CS_YUV420P10; break;
    case PixelVideo_YUV_P16: avs_vi.pixel_type=VideoInfo::CS_YUV420P16; break;
    case PixelData: avs_vi.pixel_type=VideoInfo::CS_BGR32; break;
    }
#endif
#ifdef VAPOURSYNTH_PLUGIN
    switch(type)
    {
    case PixelVideo_YV12: vs_vi.format = vsapi->getFormatPreset(pfYUV420P8, vscore); break;
    case PixelVideo_YUV_P10: vs_vi.format = vsapi->getFormatPreset(pfYUV420P10, vscore); break;
    case PixelVideo_YUV_P16: vs_vi.format = vsapi->getFormatPreset(pfYUV420P16, vscore); break;
    case PixelData: vs_vi.format = vsapi->getFormatPreset(pfGray8, vscore); break;
    default: vs_vi.format = 0;
    }
#endif
}

uint32_t BaseFilter::Video::totalFrames() const
{
#ifdef AVISYNTH_PLUGIN
    return avs_vi.num_frames;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    return vs_vi.numFrames;
#endif
}

void BaseFilter::Video::setTotalFrames(uint32_t count)
{
#ifdef AVISYNTH_PLUGIN
    avs_vi.num_frames = count;
#endif
#ifdef VAPOURSYNTH_PLUGIN
    vs_vi.numFrames = count;
#endif
}

//////////////////////////////

BaseFilter::BaseFilter():videoInfo(0),core(0)
{
}

BaseFilter::~BaseFilter()
{
    if(videoInfo)
        delete videoInfo;
}

void BaseFilter::setCore(CoreLogic *core)
{ this->core = core; }

std::string BaseFilter::getError() const
{ return errorString; }

void BaseFilter::setError(const std::string& error)
{ errorString = error; }

CoreLogic* BaseFilter::getCore()
{ return core; }

BaseFilter::Video* BaseFilter::getVideoInfo() const
{ return videoInfo; }

///////////////////////////////////////////

CoreLogic::CoreLogic(BaseFilter *base)
{
    this->base = base;
    base->setCore(this);
}

CoreLogic::~CoreLogic()
{
    delete base;
}

BaseFilter* CoreLogic::getBase() const
{ return base; }

////////////////////////////////////////////

#ifdef AVISYNTH_PLUGIN

AVSBaseFilter::AVSBaseFilter(const VideoInfo& vi,IScriptEnvironment *env):BaseFilter()
{
    mainEnv = env;
    videoInfo = new Video(vi);
    vi_inited=false;

    avsV8 = true;
    try { env->CheckVersion(8); }
    catch(const AvisynthError&) { avsV8 = false; }
}

AVSBaseFilter::~AVSBaseFilter()
{
}

PVideoFrame AVSBaseFilter::getFrame(int n,IScriptEnvironment *env)
{
    getCore()->prepare(n, env);

    std::shared_ptr<Frame> frm = getCore()->get(n, env);

    if(!frm) return 0;

    return frm->avsFrame;
}

void AVSBaseFilter::addUpstreamFilter(const std::string& name,PClip clip)
{
    if(clip) clips[name] = clip;
}

bool AVSBaseFilter::hasUpstreamFilter(const std::string& name) const
{
    return clips.find(name) != clips.end();
}

BaseFilter::Video AVSBaseFilter::getVideoInfo(const std::string& name) const
{
    return Video(clips.at(name)->GetVideoInfo());
}

BaseFilter::Video AVSBaseFilter::newVideoInfo(int w,int h,Video::PixelType type) const
{ return Video(w,h,type); }

VideoInfo AVSBaseFilter::initVideoInfo()
{
    if(!vi_inited)
    {
        vi_inited=true;
        getCore()->initVideoInfo(videoInfo);
    }
    return videoInfo->avs_vi;
}

std::shared_ptr<BaseFilter::Frame> AVSBaseFilter::newFrame(void *baseRef,std::shared_ptr<Frame> src)
{
    return newFrame(*videoInfo,baseRef,src);
}

std::shared_ptr<BaseFilter::Frame> AVSBaseFilter::newFrame(const Video& vi,void *baseRef,std::shared_ptr<Frame> src)
{
    IScriptEnvironment *env = (IScriptEnvironment*)baseRef;

    PVideoFrame avsFrame = (src && avsV8) ? env->NewVideoFrameP(vi.avs_vi,&src->avsFrame) : env->NewVideoFrame(vi.avs_vi);
    if(!avsFrame) return std::shared_ptr<Frame>(0);

    return std::shared_ptr<Frame>(new Frame(avsFrame,vi.pixelType()>=BaseFilter::Video::PixelVideo_YUV_P10 ? 2:1));
}

std::shared_ptr<BaseFilter::Frame> AVSBaseFilter::getUpstreamFrame(const std::string& filter,int n,void *baseRef)
{
    PClip clip = clips.at(filter);
    if(!clip) return std::shared_ptr<Frame>(0);

    IScriptEnvironment *env = (IScriptEnvironment*)baseRef;

    PVideoFrame avsFrame = clip->GetFrame(n,env);
    return std::shared_ptr<Frame>(new Frame(avsFrame, clip->GetVideoInfo().ComponentSize()));
}

int64_t AVSBaseFilter::getInt64(const std::string& name)
{
    if(name.front()=='#')
    {
        return (int64_t)getVideoInfo(name.substr(1)).avs_vi.num_audio_samples;
    }
    else
        try { return mainEnv->GetVar(name.c_str()).AsInt(); }
        catch(...) { return 0; }
}

void AVSBaseFilter::setInt64(const std::string& name,int64_t i)
{
    if(name=="#")
    {
        videoInfo->avs_vi.audio_samples_per_second = 0;
        videoInfo->avs_vi.nchannels = 0;
        videoInfo->avs_vi.num_audio_samples = (uint64_t)i;
    }
}

#endif

#ifdef VAPOURSYNTH_PLUGIN

VSBaseFilter::VSBaseFilter(const VSVideoInfo& vi,VSCore *vscore,const VSAPI *vsapi):BaseFilter()
{
    this->vscore = vscore;
    this->vsapi  = vsapi;

    videoInfo = new Video(vi,vsapi,vscore);
}

VSBaseFilter::~VSBaseFilter()
{
    for(auto node: nodes)
        vsapi->freeNode(node.second);
}

void VSBaseFilter::prepareFrame(int n,VSFrameContext *frameCtx)
{
    getCore()->prepare(n, frameCtx);
}

VSFrameRef* VSBaseFilter::getFrame(int n,VSFrameContext *frameCtx)
{
    std::shared_ptr<Frame> frm = getCore()->get(n, frameCtx);

    if(!frm) return 0;

	if (frm->vsFrame)
	{
		VSMap *args = vsapi->getFramePropsRW(frm->vsFrame);
		int64_t num, den;
		frm->getDuration(num, den);
        if (num >= 0 && den >= 0)
		{
			vsapi->propSetInt(args, "_DurationNum", num, paReplace);
			vsapi->propSetInt(args, "_DurationDen", den, paReplace);
		}
        vsapi->propSetInt(args, "_PTS", frm->getPTS(), paReplace);
	}

    return frm->vsFrame;
}

void VSBaseFilter::addUpstreamFilter(const std::string& name,VSNodeRef *node)
{
    if(node) nodes[name] = node;
}

bool VSBaseFilter::hasUpstreamFilter(const std::string& name) const
{
    return nodes.find(name) != nodes.end();
}

BaseFilter::Video VSBaseFilter::getVideoInfo(const std::string& name) const
{
    return Video(*(vsapi->getVideoInfo(nodes.at(name))),vsapi,vscore);
}

BaseFilter::Video VSBaseFilter::newVideoInfo(int w,int h,Video::PixelType type) const
{ return Video(w,h,type,vsapi,vscore); }

VSVideoInfo VSBaseFilter::initVideoInfo()
{
    getCore()->initVideoInfo(videoInfo);
    return videoInfo->vs_vi;
}

//BaseFilter interface
std::shared_ptr<BaseFilter::Frame> VSBaseFilter::newFrame(void* /*baseRef*/,std::shared_ptr<BaseFilter::Frame> src)
{
    return newFrame(*videoInfo,0,src);
}

std::shared_ptr<BaseFilter::Frame> VSBaseFilter::newFrame(const Video& vi, void* /*baseRef*/, std::shared_ptr<Frame> src)
{
    VSFrameRef *src_ref = src ? src->vsFrame:0;
    VSFrameRef *frm = vsapi->newVideoFrame(vi.vs_vi.format,vi.width(),vi.height(),src_ref,vscore);
    if(!frm) return std::shared_ptr<Frame>(0);

	Frame *pf = new Frame(frm, vsapi);
	if (src)
	{
		int64_t n, d;
		src->getDuration(n, d);
		pf->setDuration(n, d);
        pf->setPTS(src->getPTS());
	}

    return std::shared_ptr<Frame>(pf);
}

std::shared_ptr<BaseFilter::Frame> VSBaseFilter::copyFrame(void* /*baseRef*/,std::shared_ptr<Frame> src) const
{
    if(!src) return std::shared_ptr<Frame>(0);

    VSFrameRef *frm = vsapi->copyFrame(src->vsFrame,vscore);
    if(!frm) return std::shared_ptr<Frame>(0);

    Frame *pf = new Frame(frm, vsapi);
    if (src)
    {
        int64_t n, d;
        src->getDuration(n, d);
        pf->setDuration(n, d);
        pf->setPTS(src->getPTS());
    }

    return std::shared_ptr<Frame>(pf);
}

void VSBaseFilter::prefetchUpstreamFrame(const std::string& filter,int n,void* baseRef)
{
    VSNodeRef *node = nodes.at(filter);
    if(!node) return;

    VSFrameContext *ctx = (VSFrameContext*)baseRef;

    vsapi->requestFrameFilter(n,node,ctx);
}

std::shared_ptr<BaseFilter::Frame> VSBaseFilter::getUpstreamFrame(const std::string& filter, int n, void* baseRef)
{
    VSNodeRef *node = nodes.at(filter);
    if(!node) return std::shared_ptr<Frame>(0);

    VSFrameContext *ctx = (VSFrameContext*)baseRef;

    const VSFrameRef* frm = vsapi->getFrameFilter(n,node,ctx);
    if(!frm) return std::shared_ptr<Frame>(0);
	
	const VSMap *args = vsapi->getFramePropsRO(frm);
	int err;
	Frame *pf = new Frame(frm, vsapi);
	pf->setDuration(vsapi->propGetInt(args, "_DurationNum", 0, &err), vsapi->propGetInt(args, "_DurationDen", 0, &err));
    pf->setPTS(vsapi->propGetInt(args, "_PTS", 0, &err));

    return std::shared_ptr<Frame>(pf);
}

int64_t VSBaseFilter::getInt64(const std::string& name)
{
    if(name.front()=='#')
        return nodesParams[name.substr(1)];
    else return 0;
}

void VSBaseFilter::setInt64(const std::string& name,int64_t i)
{
    if(name.front()=='#')
        nodesParams[name.substr(1)]=i;
}

#endif
