/**************************************************************
 SVPFlow1 - motion vectors search for SmoothVideo Project (SVP)
 Based on MVTools by Manao
 Copyright(c)2011 Chainik - http://www.svp-team.com
 Copyright(c)2006 A.G.Balakhnin aka Fizick - http://avisynth.org.ru

 JsonCpp: jsoncpp.sourceforge.net under MIT

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
 http://www.gnu.org/copyleft/gpl.html
*****************************************************************/

#include <math.h>

#include "mvanalyse.h"
#include "mvsuper.h"
#include "remotecontrol.h"

#ifdef AVISYNTH_PLUGIN

static std::map<uint64_t,std::shared_ptr<MVAnalyseCore>> cores;
static std::mutex mtxInit;

MVAnalyse_AVS::MVAnalyse_AVS(PClip _child, PClip source, const JsonSVP::Value& root, IScriptEnvironment* env):GenericVideoFilter(_child)
{
    handle = uint64_t(child.operator->());

    std::lock_guard<std::mutex> lock(mtxInit);

    bool firstCall = cores.find(handle) == cores.end();
    if(firstCall)
    {
        AVSBaseFilter *avs = new AVSBaseFilter(vi,env);
        avs->addUpstreamFilter("source",source);
        avs->addUpstreamFilter("super",child);

        auto core = std::shared_ptr<MVAnalyseCore>(new MVAnalyseCore(root,avs));
        std::string err = core->getBase()->getError();
        if(!err.empty())
        {
            env->ThrowError(err.c_str());
            return;
        }

        cores[handle] = core;
    }

    core = cores.at(handle);
    vi = dynamic_cast<AVSBaseFilter*>(core->getBase())->initVideoInfo();
}

MVAnalyse_AVS::~MVAnalyse_AVS()
{
    cores.erase(handle);
}

PVideoFrame __stdcall MVAnalyse_AVS::GetFrame(int n, IScriptEnvironment* env)
{
    return dynamic_cast<AVSBaseFilter*>(core->getBase())->getFrame(n,env);
}

#endif //AVISYNTH_PLUGIN

#ifdef VAPOURSYNTH_PLUGIN

void VS_CC MVAnalyse_VS::create(const VSMap *in, VSMap *out, void* /*userData*/, VSCore *vscore, const VSAPI *vsapi)
{
    const char *opt = vsapi->propGetData(in, "opt", 0,0);
    int len = vsapi->propGetDataSize(in, "opt", 0,0);

    JsonSVP::Value root;
    JsonSVP::Reader reader;
    if(len>0 && !reader.parse(opt,opt+len,root))
    {
        vsapi->setError(out,("SVAnalyse: invalid 'params' syntax: "+reader.getFormatedErrorMessages()).c_str());
        return;
    }

    VSNodeRef *node = vsapi->propGetNode(in, "clip", 0, 0);
    VSBaseFilter *base = new VSBaseFilter(*(vsapi->getVideoInfo(node)),vscore,vsapi);

    base->addUpstreamFilter("super",node);
    base->addUpstreamFilter("source",vsapi->propGetNode(in, "src", 0, 0));

    base->setInt64("#super",vsapi->propGetInt(in,"sdata",0,0));

    MVAnalyseCore *core = new MVAnalyseCore(root,base);
    std::string err = core->getBase()->getError();
    if(!err.empty())
    {
        vsapi->setError(out,err.c_str());

        delete core;
        return;
    }

    vsapi->propSetInt(out,"data",base->getInt64("#"),0);
    vsapi->createFilter(in,out,"SVAnalyse",&MVAnalyse_VS::init,&MVAnalyse_VS::getFrame,&MVAnalyse_VS::free,fmParallel,0,core,vscore);
}

void VS_CC MVAnalyse_VS::init(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore* /*vscore*/, const VSAPI *vsapi)
{
    MVAnalyseCore *core = static_cast<MVAnalyseCore*>(*instanceData);

    VSVideoInfo vi = dynamic_cast<VSBaseFilter*>(core->getBase())->initVideoInfo();
    vsapi->setVideoInfo(&vi, 1, node);
}

const VSFrameRef* VS_CC MVAnalyse_VS::getFrame(int n, int activationReason, void **instanceData, void** /*frameData*/, VSFrameContext *frameCtx, VSCore* /*vscore*/, const VSAPI* /*vsapi*/)
{
    MVAnalyseCore *core = static_cast<MVAnalyseCore*>(*instanceData);

    if(activationReason==arInitial)
        dynamic_cast<VSBaseFilter*>(core->getBase())->prepareFrame(n,frameCtx);
    else if(activationReason==arAllFramesReady)
        return dynamic_cast<VSBaseFilter*>(core->getBase())->getFrame(n,frameCtx);

    return 0;
}

void VS_CC MVAnalyse_VS::free(void *instanceData, VSCore* /*vscore*/, const VSAPI* /*vsapi*/)
{
    MVAnalyseCore *core = static_cast<MVAnalyseCore*>(instanceData);
    delete core;
}

#endif

static SearchType _intToSearchType(int t,SearchType def)
{
    SearchType res=def;
    switch(t)
    {
    case 2:	res = HEX2SEARCH; break;
    case 3:	res = UMHSEARCH; break;
    default: res = EXHAUSTIVE; break;
    }
    return res;
}

#define SET_ERROR(str) {getBase()->setError(str); return;}

MVAnalyseCore::MVAnalyseCore(const JsonSVP::Value& root, BaseFilter *base):CoreLogic(base)
{
    BaseFilter::Video *vi = base->getVideoInfo();
    if (vi->pixelType() != BaseFilter::Video::PixelVideo_YV12)
        SET_ERROR("SVAnalyse: Clip must be YV12");

    uint64_t p = base->getInt64("#super");
    memcpy(&superParams, &p, 8);

    if (superParams.nHeight <=0 || superParams.nPel<1 || superParams.nPel>4 || superParams.nLevels<1)
        SET_ERROR("SVAnalyse: wrong super clip (pseudoaudio) parameters");

    if(superParams.flags&SUPER_NOFINESTLEVEL && !base->hasUpstreamFilter("source"))
        SET_ERROR("SVAnalyse: source clip needed for 'optimized' superclip");

    hasFinestLevel=(superParams.flags&SUPER_NOFINESTLEVEL ? false:true);

    //MSuper from MVTools can be used here
    int gpu=root.get("gpu",-1).asInt();
    bool superFromMVTools=(superParams.flags>3 || superParams.gpuMode>2 || gpu>=0);

    bool remoteControl = ((superParams.flags&SUPER_REMOTECONTROL && !superFromMVTools)?true:false);
    ad.gpuMode=(!superFromMVTools ? superParams.gpuMode : (gpu<0 ? 0 : gpu));

    ad.nWidth = superParams.nWidth;
    ad.nHeight = superParams.nHeight;

    ad.vectorsType=root.get("vectors",3).asUInt();
    if(ad.vectorsType<1 || ad.vectorsType>3)
        SET_ERROR("SVAnalyse: vectors type must be between 1 and 3");

    ad.nPel = superParams.nPel;
    if (( ad.nPel != 1 ) && ( ad.nPel != 2 ) && ( ad.nPel != 4 ))
        SET_ERROR("SVAnalyse: pel must be 1 or 2 or 4");

    ad.nBlkSizeX = root["block"].get("w",16).asUInt();
    ad.nBlkSizeY = root["block"].get("h",ad.nBlkSizeX).asUInt();
    int overlap = root["block"].get("overlap",2).asInt();
    switch(overlap)
    {
    case 0: break;
    case 1: overlap=8; break;
    case 2: overlap=4; break;
    case 3: overlap=2; break;
    default: SET_ERROR("SVAnalyse: overlap must be 0, 1 (=1/8*block), 2 (=1/4*block) or 3 (=1/2*block)");
    }
    ad.nOverlapX = (overlap==0 ? 0 : ad.nBlkSizeX/overlap);
    ad.nOverlapY = (overlap==0 ? 0 : ad.nBlkSizeY/overlap);

    if (    ( ad.nBlkSizeX != 4 || ad.nBlkSizeY != 4 ) &&
            ( ad.nBlkSizeX != 8 || ad.nBlkSizeY != 4) &&
            ( ad.nBlkSizeX != 8 || ad.nBlkSizeY != 8 ) &&
            ( ad.nBlkSizeX != 16 || ad.nBlkSizeY != 8 ) &&
            ( ad.nBlkSizeX != 16 || ad.nBlkSizeY != 16 ) &&
            ( ad.nBlkSizeX != 32 || ad.nBlkSizeY != 32 ) &&
            ( ad.nBlkSizeX != 32 || ad.nBlkSizeY != 16))
        SET_ERROR("SVAnalyse: Block's size must be 8x8, 16x8, 16x16, 32x16, 32x32");

    if(ad.nBlkSizeX==4)
        ad.nOverlapX=ad.nOverlapY=0;

//	if (ad.nOverlapX%2 || ad.nOverlapY%2)
//		env->ThrowError("SVAnalyse: Overlap must be >= 2px");

    ad.nBlkX = (ad.nWidth - ad.nOverlapX) / (ad.nBlkSizeX - ad.nOverlapX);
    ad.nBlkY = (ad.nHeight - ad.nOverlapY) / (ad.nBlkSizeY - ad.nOverlapY);

    int nWidth_B = (ad.nBlkSizeX - ad.nOverlapX)*ad.nBlkX + ad.nOverlapX; // covered by blocks
    int nHeight_B = (ad.nBlkSizeY - ad.nOverlapY)*ad.nBlkY + ad.nOverlapY;
    // calculate valid levels
    int nLevelsMax = 0;
    while (	((nWidth_B>>nLevelsMax) - ad.nOverlapX)/(ad.nBlkSizeX-ad.nOverlapX) > 0 &&
            ((nHeight_B>>nLevelsMax) -ad.nOverlapY)/(ad.nBlkSizeY-ad.nOverlapY) > 0) // at last one block
    {
        nLevelsMax++;
    }

    int lv=root["main"].get("levels",0).asInt();
    ad.nLvCount = (lv > 0 ? lv :  nLevelsMax + lv);
    if (ad.nLvCount > superParams.nLevels)
        SET_ERROR("SVAnalyse: it is not enough levels  in super clip");// (%d), while SVAnalyse asks %d", nSuperLevels, ad.nLvCount);
    if ( ad.nLvCount < 2 || ad.nLvCount > nLevelsMax)
        SET_ERROR("SVAnalyse: non-valid number of levels");// (%d)", ad.nLvCount);

    int blkMul=ad.nBlkSizeX*ad.nBlkSizeY, blkDiv=64;
    if(CHROMA_SHIFT==1) {blkMul*=4; blkDiv*=3;}
    else if(CHROMA_SHIFT==2) {blkMul*=2;}

    double lambda_in=root["main"]["penalty"].get("lambda",10.0).asDouble();
    int nLambda = int(lambda_in*1000.0*blkMul/blkDiv);
    int lsad_in=root["main"]["penalty"].get("lsad",8000).asInt();
    int lsad = lsad_in*blkMul/blkDiv;
    plevel = root["main"]["penalty"].get("plevel",1.5).asDouble();
    if(plevel<=0.01) plevel=1.0;

    int pnew = root["main"]["penalty"].get("pnew",50).asInt();
    pglobal = root["main"]["penalty"].get("pglobal",50).asInt();
    pzero = root["main"]["penalty"].get("pzero",100).asInt();
    pnbour = root["main"]["penalty"].get("pnbour",50).asInt();
    preverse = root["main"]["penalty"].get("prev",0).asInt();

    int badSAD = root["main"]["search"]["coarse"]["bad"].get("sad",1000).asInt()*blkMul/blkDiv;
    badrange = root["main"]["search"]["coarse"]["bad"].get("range",-24).asInt();

    sorting = root["main"]["search"].get("sort",true).asBool();
    tryMany = root["main"]["search"]["coarse"].get("trymany",false).asBool();

    searchType=_intToSearchType(root["main"]["search"].get("type",4).asInt(),EXHAUSTIVE);
    SearchType coarseSearchType=_intToSearchType(root["main"]["search"]["coarse"].get("type",4).asInt(),EXHAUSTIVE);

    searchParam = root["main"]["search"].get("distance",-2*ad.nPel).asInt();
    int coarseSearchParam=root["main"]["search"]["coarse"].get("distance",0).asInt();

    bool useSATD=root["main"]["search"].get("satd",false).asBool();
    bool useSATD_coarse=root["main"]["search"]["coarse"].get("satd",true).asBool();

    int coarseWidth = root["main"]["search"]["coarse"].get("width",1050).asInt();

    ad.delta = _fmax(1, root["special"].get("delta", 1).asInt());

    int flags=root.get("flags",0).asInt();

    GroupOfPlanes::InitData init = {ad,coarseSearchType,coarseSearchParam,nLevelsMax-ad.nLvCount,
                                    nLambda,lsad,pnew,(useSATD ? 1 : 0) + (useSATD_coarse ? 2 : 0), badSAD, coarseWidth, flags};

    vfData.push_back(init);

    int sz=root["refine"].size();
    if(ad.gpuMode==2) sz=0; //no refine with GPU acceleration
    if(ad.nBlkSizeX==4) sz=0;
    for(int i=0;i<sz;i++)
    {
        ad.nLvCount = 1;

        const JsonSVP::Value& v=root["refine"][i];
        bool divide=true;
        if(divide)
        {
            if(ad.nOverlapX==1 || ad.nOverlapY==1)
                SET_ERROR("SVAnalyse: Overlap size must be >= 1 after recalcualtion");
            if(ad.nBlkSizeX<8 || ad.nBlkSizeY<8)
                SET_ERROR("SVAnalyse: Block's size must be >= 4x4 after recalcualtion");

            ad.nBlkSizeX/=2;
            ad.nBlkSizeY/=2;
            ad.nOverlapX/=2;
            ad.nOverlapY/=2;
            ad.nBlkX = (ad.nWidth - ad.nOverlapX) / (ad.nBlkSizeX - ad.nOverlapX);
            ad.nBlkY = (ad.nHeight - ad.nOverlapY) / (ad.nBlkSizeY - ad.nOverlapY);

            blkMul/=4;
            lsad/=4;
        }

        int recalcSAD = v.get("thsad",200).asInt()*blkMul/blkDiv;
        if(recalcSAD<0) recalcSAD = INT_MAX;
        bool satd=v["search"].get("satd",useSATD).asBool();

        GroupOfPlanes::InitData init = {ad,
                                        _intToSearchType(v["search"].get("type",(int)searchType).asInt(),EXHAUSTIVE),
                                        _fmax(1, v["search"].get("distance", ad.nPel).asInt()),
                                        0,0,lsad,
                                        v["penalty"].get("pnew",pnew).asInt(),
                                        satd?1:0,
                                        recalcSAD,0,flags};

        vfData.push_back(init);
    }

    headerSize = 1+sizeof(SVAnalysisData)/sizeof(int);
    totalVectorsCount = 1 + 2*ad.nBlkX*ad.nBlkY;

    base->setInt64("#",int64_t(&ad));

#ifndef Q_OS_MAC
    if(remoteControl)
        RemoteControl::create((uint64_t)this);
#endif
}

MVAnalyseCore::~MVAnalyseCore()
{
    RemoteControl::remove((uint64_t)this);
}

void MVAnalyseCore::initVideoInfo(BaseFilter::Video *vi)
{
    vi->setPixelType(BaseFilter::Video::PixelData);
    vi->setSize(4 * (headerSize + 2 * totalVectorsCount), 1);
}

#define CHECK_NUM_ARGS(n) if(it->args.size()!=(n)) continue
#define CHECK_ARG(n) (it->args[n]!=IGNORE_VALUE)

void MVAnalyseCore::checkRemoteMessages()
{
    RemoteControl::MessageList list=RemoteControl::exchange((uint64_t)this);
    if(!list.size()) return;

    int blkMul=ad.nBlkSizeX*ad.nBlkSizeY, blkDiv=64;
    if(CHROMA_SHIFT==1) {blkMul*=4; blkDiv*=3;}
    else if(CHROMA_SHIFT==2) {blkMul*=2;}

    GroupOfPlanes::PARAMS_CHANGE pc;
    RemoteControl::MessageList::iterator it;
    for(it=list.begin();it!=list.end();++it)
        switch(it->id)
        {
        case 21: //SVP-shader
            CHECK_NUM_ARGS(2);
            ad.vectorsType=VECTORS_BACKWARD + (it->args[0]==1 ? 0:VECTORS_FORWARD);
            pc.vectorsType=ad.vectorsType;
            for(int i=0;i<(int)vfData.size();i++)
                vfData[i].changeParams(pc);
            break;

        case 101: {//badsad
            CHECK_NUM_ARGS(2);
            if(CHECK_ARG(0))
            {
                pc.thSAD=it->args[0]*blkMul/blkDiv;
                vfData[0].changeParams(pc);
            }
            if(CHECK_ARG(1)) badrange=it->args[1];
            } break;

        case 102: { //search type & distance
            CHECK_NUM_ARGS(4);
            if(CHECK_ARG(0)) searchType=_intToSearchType(it->args[0],EXHAUSTIVE);
//			if(CHECK_ARG(1)) searchParam = max(0,it->args[1]);
            if(CHECK_ARG(1)) searchParam = it->args[1];

            pc.searchType=(CHECK_ARG(2) ? _intToSearchType(it->args[2],EXHAUSTIVE) : IGNORE_VALUE);
            pc.nSearchParam=it->args[3];
            vfData[0].changeParams(pc);

            pc.searchType=searchType;
            pc.nSearchParam=searchParam;
            for(size_t i=1;i<vfData.size();i++)
                vfData[i].changeParams(pc);
            } break;

        case 103: { //satd
            CHECK_NUM_ARGS(2);
            if(CHECK_ARG(0)) pc.satd=it->args[0];
            if(CHECK_ARG(1)) pc.satdCoarse=it->args[2];
            for(size_t i=0;i<vfData.size();i++)
                vfData[i].changeParams(pc);
            } break;

        case 110: { //lambda
            CHECK_NUM_ARGS(3);
            if(CHECK_ARG(0)) pc.nLambda=int(it->args[0]/100.0*1000*blkMul/blkDiv);
            if(CHECK_ARG(1)) plevel=it->args[1]/100.0;
            if(CHECK_ARG(2)) pc.lsad=it->args[2]*blkMul/blkDiv;
            for(size_t i=0;i<vfData.size();i++)
            {
                vfData[i].changeParams(pc);
                pc.lsad/=4;
            }
            } break;

        case 111: { //penalty
            CHECK_NUM_ARGS(5);
            if(CHECK_ARG(0))
            {
                pc.pnew=it->args[0];
                for(size_t i=0;i<vfData.size();i++)
                    vfData[i].changeParams(pc);
            }
            if(CHECK_ARG(1)) pglobal=it->args[1];
            if(CHECK_ARG(2)) pzero=it->args[2];
            if(CHECK_ARG(3)) pnbour=it->args[3];
            if(CHECK_ARG(4)) preverse=it->args[4];
            } break;

        case 115: { //thsad for ALL (?) refinemens
            CHECK_NUM_ARGS(1);
            if(CHECK_ARG(0) && vfData.size()>1)
                pc.thSAD=it->args[0]*blkMul/blkDiv;
            if(pc.thSAD<0) pc.thSAD = INT_MAX;
            for(size_t i=1;i<vfData.size();i++)
                vfData[i].changeParams(pc);
            } break;

        case 120: { //debug (?)
            CHECK_NUM_ARGS(3);
            if(CHECK_ARG(0)) sorting=(it->args[0]?true:false);
            if(CHECK_ARG(1)) tryMany=(it->args[1]?true:false);
            if(CHECK_ARG(2))
            {
                pc.flags=it->args[2];
                for(int i=0;i<(int)vfData.size();i++)
                    vfData[i].changeParams(pc);
            }
            } break;
        }
}

void MVAnalyseCore::prepare(int n, void *baseRef)
{
    getBase()->prefetchUpstreamFrame("super",n,baseRef);
    getBase()->prefetchUpstreamFrame("super",n+ad.delta,baseRef);

    if(!hasFinestLevel)
    {
        getBase()->prefetchUpstreamFrame("source",n,baseRef);
        getBase()->prefetchUpstreamFrame("source",n+ad.delta,baseRef);
    }
}

std::shared_ptr<BaseFilter::Frame> MVAnalyseCore::get(int n, void *baseRef)
{
    checkRemoteMessages();

    auto dst = getBase()->newFrame(baseRef);
    uint8_t *pDst1;
    dst->write(&pDst1,0,0);

    bool valid=(n>=0 && (uint32_t)n<getBase()->getVideoInfo()->totalFrames());
    if(!valid) ad.vectorsType=0;

    // write analysis parameters as a header to frame
    memcpy(pDst1, &headerSize, sizeof(int));
    memcpy(pDst1+sizeof(int), &ad, sizeof(SVAnalysisData));
    pDst1 += headerSize*sizeof(int);
    uint8_t *pDst2 = pDst1 + totalVectorsCount*sizeof(int);

    if(!valid) return dst;

    MVGroupOfFrames *pFrame1GOF = new MVGroupOfFrames(superParams.nLevels, ad.nWidth, ad.nHeight, ad.nPel, hasFinestLevel);
    MVGroupOfFrames *pFrame2GOF = new MVGroupOfFrames(superParams.nLevels, ad.nWidth, ad.nHeight, ad.nPel, hasFinestLevel);

    std::vector<GroupOfPlanes*> vectorFields;
    for(GroupOfPlanes::InitData d: vfData)
        vectorFields.push_back(new GroupOfPlanes(d));

    auto frame1 = getBase()->getUpstreamFrame("super",n,baseRef);
    auto frame2 = getBase()->getUpstreamFrame("super",n+ad.delta,baseRef);

    uint8_t *y, *u, *v;
    frame1->read(&y, &u, &v);
    pFrame1GOF->Update(y,frame1->pitchY(),u,v,frame1->pitchUV());
    frame2->read(&y, &u, &v);
    pFrame2GOF->Update(y,frame2->pitchY(),u,v,frame2->pitchUV());

    std::shared_ptr<BaseFilter::Frame> srcFrame1,srcFrame2;
    if(!hasFinestLevel)
    {
        srcFrame1 = getBase()->getUpstreamFrame("source",n,baseRef);
        srcFrame2 = getBase()->getUpstreamFrame("source",n+ad.delta,baseRef);

        srcFrame1->read(&y, &u, &v);
        pFrame1GOF->SetPlanes(y,srcFrame1->pitchY(),u,v,srcFrame1->pitchUV());
        srcFrame2->read(&y, &u, &v);
        pFrame2GOF->SetPlanes(y,srcFrame2->pitchY(),u,v,srcFrame2->pitchUV());
    }

    int sz=(int)vectorFields.size();

    bool backOnly=(ad.vectorsType==VECTORS_BACKWARD);
    int *dst1 = (int*)pDst1;
    int *dst2 = (int*)pDst2;

    int lambdaOut;
    vectorFields[0]->SearchMVs(backOnly ? pFrame1GOF:pFrame2GOF, backOnly ? pFrame2GOF:pFrame1GOF,
                                    searchType, searchParam, plevel,
                                    pzero, pglobal, pnbour, preverse, badrange, tryMany, sorting,
                                    &lambdaOut,
                                    sz>1 ? 0 : (backOnly ? dst2 : dst1), sz>1 ? 0 : dst2);

    for(int i=1;i<sz;i++)
        vectorFields[i]->RecalculateMVs(vectorFields[i-1],
                                    backOnly ? pFrame1GOF:pFrame2GOF, backOnly ? pFrame2GOF:pFrame1GOF,
                                    &lambdaOut,
                                    i==sz-1 ? (backOnly ? dst2 : dst1) : 0, i==sz-1 ? dst2 : 0);

    for(int i=0;i<(int)vectorFields.size();i++)
        delete vectorFields[i];

    delete pFrame1GOF;
    delete pFrame2GOF; 

    _clear_sse();

	return dst;
}
