/**************************************************************
 SVPFlow1 - motion vectors search for SmoothVideo Project (SVP)
 Based on MVTools by Manao
 Copyright(c)2011 Chainik - http://www.svp-team.com
 Copyright(c)2006 A.G.Balakhnin aka Fizick - http://avisynth.org.ru

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
 http://www.gnu.org/copyleft/gpl.html
*****************************************************************/

#include <math.h>

#include "remotecontrol.h"
#include "mvsuper.h"

#ifdef AVISYNTH_PLUGIN

static std::map<uint64_t,std::shared_ptr<MVSuperCore>> cores;
static std::mutex mtxInit;

MVSuper_AVS::MVSuper_AVS(PClip _child, const JsonSVP::Value& root, IScriptEnvironment* env):GenericVideoFilter(_child)
{
    handle = uint64_t(child.operator->());

    std::lock_guard<std::mutex> lock(mtxInit);

    bool firstCall = cores.find(handle) == cores.end();
    if(firstCall)
    {
        AVSBaseFilter *avs = new AVSBaseFilter(vi,env);
        avs->addUpstreamFilter("source",child);

        auto core = std::shared_ptr<MVSuperCore>(new MVSuperCore(root,avs));
        std::string err = core->getBase()->getError();
        if(!err.empty())
        {
            env->ThrowError(err.c_str());
            return;
        }

        cores[handle] = core;
    }

    core = cores.at(handle);
    vi = dynamic_cast<AVSBaseFilter*>(core->getBase())->initVideoInfo();
}

MVSuper_AVS::~MVSuper_AVS()
{
    cores.erase(handle);
}

PVideoFrame __stdcall MVSuper_AVS::GetFrame(int n, IScriptEnvironment* env)
{
    return dynamic_cast<AVSBaseFilter*>(core->getBase())->getFrame(n,env);
}

#endif //AVISYNTH_PLUGIN

#ifdef VAPOURSYNTH_PLUGIN

void VS_CC MVSuper_VS::create(const VSMap *in, VSMap *out, void* /*userData*/, VSCore *vscore, const VSAPI *vsapi)
{
    const char *opt = vsapi->propGetData(in, "opt", 0,0);
    int len = vsapi->propGetDataSize(in, "opt", 0,0);

    JsonSVP::Value root;
    JsonSVP::Reader reader;
    if(len>0 && !reader.parse(opt,opt+len,root))
    {
        vsapi->setError(out,("SVSuper: invalid 'params' syntax: "+reader.getFormatedErrorMessages()).c_str());
        return;
    }

    VSNodeRef *node = vsapi->propGetNode(in, "clip", 0, 0);

    VSBaseFilter *base = new VSBaseFilter(*(vsapi->getVideoInfo(node)),vscore,vsapi);
    base->addUpstreamFilter("source",node);

    MVSuperCore *core = new MVSuperCore(root,base);
    std::string err = core->getBase()->getError();
    if(!err.empty())
    {
        vsapi->setError(out,err.c_str());

        delete core;
        return;
    }

    vsapi->propSetInt(out,"data",base->getInt64("#"),0);
    vsapi->createFilter(in,out,"SVSuper",&MVSuper_VS::init,&MVSuper_VS::getFrame,&MVSuper_VS::free,fmParallel,0,core,vscore);
}

void VS_CC MVSuper_VS::init(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore* /*vscore*/, const VSAPI *vsapi)
{
    MVSuperCore *core = static_cast<MVSuperCore*>(*instanceData);

    VSVideoInfo vi = dynamic_cast<VSBaseFilter*>(core->getBase())->initVideoInfo();
    vsapi->setVideoInfo(&vi, 1, node);
}

const VSFrameRef* VS_CC MVSuper_VS::getFrame(int n, int activationReason, void **instanceData, void** /*frameData*/, VSFrameContext *frameCtx, VSCore* /*vscore*/, const VSAPI* /*vsapi*/)
{
    MVSuperCore *core = static_cast<MVSuperCore*>(*instanceData);

    if(activationReason==arInitial)
        dynamic_cast<VSBaseFilter*>(core->getBase())->prepareFrame(n,frameCtx);
    else if(activationReason==arAllFramesReady)
        return dynamic_cast<VSBaseFilter*>(core->getBase())->getFrame(n,frameCtx);

    return 0;
}

void VS_CC MVSuper_VS::free(void *instanceData, VSCore* /*vscore*/, const VSAPI* /*vsapi*/)
{
    MVSuperCore *core = static_cast<MVSuperCore*>(instanceData);
    delete core;
}

#endif

#define SET_ERROR(str) {getBase()->setError(str); return;}

MVSuperCore::MVSuperCore(const JsonSVP::Value& root, BaseFilter *base):CoreLogic(base)
{
    memset(&params, 0, 8);

    BaseFilter::Video *vi = base->getVideoInfo();
    if (vi->pixelType() != BaseFilter::Video::PixelVideo_YV12)
        SET_ERROR("SVSuper: Clip must be YV12");

    int nPel = root.get("pel",2).asUInt();
    if (( nPel != 1 ) && ( nPel != 2 ) && ( nPel != 4 ))
        SET_ERROR("SVSuper: pel has to be 1 or 2 or 4");

    gpuMode = root.get("gpu",1).asUInt();
    if(gpuMode>2)
        SET_ERROR("SVSuper: gpu has to be 0 or 1 or 2");

    rfilter = root["scale"].get("down",4).asUInt();
    sharp = root["scale"].get("up",2).asUInt();

    needFinestLevel = root.get("full",true).asBool();
    if(!needFinestLevel && nPel>1)
        SET_ERROR("SVSuper: can't ignore finest level with high pel value");

    if(gpuMode) sharp=0;

    int nLevelsMax = 0;
    while (	planeSize(vi->height(), nLevelsMax) >= 2*2 &&
            planeSize(vi->width(),  nLevelsMax) >= 2*2) // at last two pixels width and height of chroma
        nLevelsMax++;

    int nLevels = nLevelsMax;

    bool remoteControl = root.get("rc",false).asBool();
#ifndef Q_OS_MAC
    if(remoteControl)
        RemoteControl::create((uint64_t)this);
#endif

    params.nWidth = vi->width();
    params.nHeight = vi->height();
    params.nPel = nPel;
    params.nLevels = nLevels;
    params.gpuMode = gpuMode;
    params.flags=0;
    if(remoteControl) params.flags|=SUPER_REMOTECONTROL;
    if(!needFinestLevel) params.flags|=SUPER_NOFINESTLEVEL;

    base->setInt64("#",*((int64_t*)&params));
}

MVSuperCore::~MVSuperCore()
{
    RemoteControl::remove((uint64_t)this);
}

void MVSuperCore::initVideoInfo(BaseFilter::Video *vi)
{
    int w = vi->width(), h=vi->height();

    if(!needFinestLevel)
        w = planeSize(w,1);

    h = planeSuperOffset(false, h, params.nLevels, params.nPel, w, needFinestLevel)/w;
    if(!needFinestLevel)
    {
        int sz = vi->width()*vi->height()/2;
        h = _fmax(h, (int)ceil((double)sz / w));
    }

    if(h&1) h++;

    vi->setSize(w,h);
}

int MVSuperCore::planeSize(int src_size, int level)
{
    int sz = src_size;

    for (int i=1; i<=level; i++)
        sz = ((sz/2) / 2) * 2;
    return sz;
}

unsigned int MVSuperCore::planeSuperOffset(bool chroma, int src_height, int level, int pel, int plane_pitch, bool hasFinestLevel)
{
    int height = src_height;
    unsigned int offset;

    if (level==0)
        offset = 0;
    else
    {
        offset = (hasFinestLevel ? pel*pel*plane_pitch*src_height : 0);

        for (int i=1; i<level; i++)
        {
            height = planeSize(src_height*(chroma?2:1), i);
            if(chroma) height/=2;

            offset += plane_pitch*height;
        }
    }
    return offset;
}

void MVSuperCore::checkRemoteMessages()
{
    RemoteControl::MessageList list=RemoteControl::exchange((uint64_t)this);
    if(!list.size()) return;

    RemoteControl::MessageList::iterator it;
    for(it=list.begin();it!=list.end();++it)
        switch(it->id)
        {
        case 201:
            if(it->args.size()!=2) continue;
            if (it->args[0] != IGNORE_VALUE) sharp = _fmin(2, _fmax(0, it->args[0]));
            if (it->args[1] != IGNORE_VALUE) rfilter = _fmin(4, _fmax(0, it->args[1]));
            break;
        }
}

void MVSuperCore::prepare(int n, void *baseRef)
{
    getBase()->prefetchUpstreamFrame("source",n,baseRef);
}

std::shared_ptr<BaseFilter::Frame> MVSuperCore::get(int n, void *baseRef)
{
    checkRemoteMessages();

    auto src = getBase()->getUpstreamFrame("source", n,baseRef);
    auto dst = getBase()->newFrame(baseRef,src);

    auto pSrcGOF = new MVGroupOfFrames(params.nLevels, params.nWidth, params.nHeight, params.nPel, needFinestLevel);

    uint8_t *y, *u, *v;
    dst->write(&y,&u,&v);
    pSrcGOF->Update(y, dst->pitchY(), u, v, dst->pitchUV());

    src->read(&y, &u, &v);
    pSrcGOF->SetPlanes(y, src->pitchY(), u, v, src->pitchUV());

    pSrcGOF->Reduce(rfilter);

    if(gpuMode<2)
        pSrcGOF->Refine(sharp);

    _clear_sse();

    delete pSrcGOF;

    return dst;
}
