/**************************************************************
 SVPFlow1 - motion vectors search for SmoothVideo Project (SVP)
 Copyright(c)2011 Chainik - http://www.svp-team.com

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
 http://www.gnu.org/copyleft/gpl.html
*****************************************************************/

#include <mutex>
#include "remotecontrol.h"

#ifdef Q_OS_WIN
#define MANAGER_PIPE L"\\\\.\\pipe\\svp-team.com.manager4"
#else
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdlib.h>
#define MANAGER_PIPE "svp-team.com.manager4"
#define INVALID_HANDLE_VALUE -1
#endif

static RemoteControl *remoteControl=0;
static std::mutex rcMtx;

int RemoteControl::Message::toByte(uint8_t* buf) const
{
    buf[0] = id;
    buf[1] = (uint8_t)args.size();
    int n = raw.size();
    memcpy(buf + 2, &n, 4);

    for (unsigned int i = 0; i<args.size(); i++)
    {
        n = args[i];
        memcpy(buf + 6 + i * 4, &n, 4);
    }
    if (raw.size())
        memcpy(buf + 6 + 4 * (int)args.size(), raw.data(), raw.size());

    return 6 + 4 * (int)args.size() + raw.size();
}

int RemoteControl::Message::fromByte(const uint8_t* buf, int len)
{
    if (len<6) return 0;

    int data_sz = *((int*)(buf + 2));
    int sz = 6 + buf[1] * 4 + data_sz;
    if (len<sz) return 0;

    id = buf[0];
    args.clear();
    for (int i = 0; i < buf[1]; i++)
        args.push_back(*((int*)&buf[6 + i * 4]));

    raw.resize(data_sz);
    if (data_sz)
        memcpy(raw.data(), buf + 6 + 4 * args.size(), data_sz);

    return sz;
}

void RemoteControl::create(uint64_t client)
{
    std::lock_guard<std::mutex> lock(rcMtx);
    if(!remoteControl)
        remoteControl=new RemoteControl();

    remoteControl->messages[client].clear();
}

void RemoteControl::remove(uint64_t client)
{
    std::lock_guard<std::mutex> lock(rcMtx);
    if(!remoteControl) return;

    remoteControl->messages.erase(client);
    if(!remoteControl->messages.size())
    {
        delete remoteControl;
        remoteControl=0;
    }
}

RemoteControl::MessageList RemoteControl::exchange(uint64_t client, const MessageList &list)
{
    std::lock_guard<std::mutex> lock(rcMtx);
    RemoteControl::MessageList res;
    if(remoteControl)
        res = remoteControl->exchange_p(client, list);
    return res;
}

RemoteControl::RemoteControl()
{
    hPipe = INVALID_HANDLE_VALUE;
    lastConnectTry -= std::chrono::minutes(1);

    tryToConnect();
}

RemoteControl::~RemoteControl()
{
#ifdef Q_OS_WIN
    CloseHandle(hPipe);
#else
    if(hPipe>0) ::close(hPipe);
#endif
}

bool RemoteControl::tryToConnect()
{
    auto now = std::chrono::steady_clock::now();
    if(std::chrono::duration_cast<std::chrono::milliseconds>(now-lastConnectTry).count() < 1000)
        return false;
    lastConnectTry = now;

#ifdef Q_OS_WIN
    hPipe = CreateFileW(MANAGER_PIPE, PIPE_ACCESS_DUPLEX, 0, NULL, OPEN_EXISTING, 0, NULL);
    if (hPipe == INVALID_HANDLE_VALUE) return false;
#else
#   ifdef Q_OS_MAC
    std::string path(getenv("TMPDIR"));
#   else
    std::string path("/tmp/");
#   endif
    path += std::string(MANAGER_PIPE);

    hPipe = ::socket(AF_UNIX, SOCK_STREAM, 0);
    if(hPipe<0) return false;

    sockaddr_un server;
#ifdef Q_OS_MAC
    server.sun_len=sizeof(server);
#endif
    server.sun_family = AF_UNIX;
    strcpy(server.sun_path, path.c_str());

    if(connect(hPipe, (sockaddr*)&server, sizeof(sockaddr_un)) < 0)
    {
        ::close(hPipe);
        hPipe = -1;
        return false;
    }
#endif

    uint8_t buf[1024];
    RemoteControl::Message msg;
    msg.id = 0;
    msg.args.push_back(1);
#ifdef Q_OS_WIN
    msg.args.push_back((int)GetCurrentProcessId());
#else
    msg.args.push_back((int)getpid());
#endif
    uint32_t written = msg.toByte(buf);
#ifdef Q_OS_WIN
    WriteFile(hPipe, buf, written, (DWORD*)&written, NULL);
#else
    int snd = ::send(hPipe,buf, written, 0);
    if(snd<0)
    {
        ::close(hPipe);
        hPipe=-1;
    }
#endif

    return true;
}

RemoteControl::MessageList RemoteControl::exchange_p(uint64_t client, const MessageList& out)
{
    RemoteControl::MessageList list;

	if (hPipe == INVALID_HANDLE_VALUE && !tryToConnect())
        return list;

    RemoteControl::Message msg;
    uint8_t buf[10240];

#ifdef Q_OS_WIN
    while (1)
    {
        DWORD avail = 0;
        if (!PeekNamedPipe(hPipe, 0, 0, 0, &avail, 0))
        {
            if (GetLastError() != ERROR_SUCCESS) //ERROR_PIPE_NOT_CONNECTED, ERROR_BROKEN_PIPE ...
            {
                CloseHandle(hPipe);
                lastConnectTry -= std::chrono::minutes(1);
                hPipe = INVALID_HANDLE_VALUE;
                return list;
            }
        }
        if (!avail) break;

        if (ReadFile(hPipe, buf, 10240, &avail, NULL))
        {
            if (!msg.fromByte(buf, avail)) continue;

            std::map<uint64_t, MessageList>::iterator it;
            for (it = messages.begin(); it != messages.end(); ++it)
                it->second.push_back(msg);
        }
    }
#else
    fd_set set;
    FD_ZERO(&set);
    FD_SET(hPipe, &set);
    timeval timeout={0,0};
    int n = ::select(hPipe+1, &set, 0,0,&timeout);
    if(n>0 && FD_ISSET(hPipe, &set))
    {
        int avail = ::recv(hPipe,(char*)buf,10240,0), sz;
        while(avail>0 && (sz=msg.fromByte(buf,avail)))
        {
            for (auto it = messages.begin(); it != messages.end(); ++it)
                it->second.push_back(msg);

            avail-=sz;
        }
    }
    else if(n == -1)
    {
        ::close(remoteControl->hPipe);
        remoteControl->lastConnectTry -= std::chrono::minutes(1);
        remoteControl->hPipe = -1;
        return list;
    }
#endif

    if (out.size())
    {
        MessageList::const_iterator it;
        for (it = out.begin(); it != out.end(); ++it)
        {
            uint32_t written = it->toByte(buf);
#ifdef Q_OS_WIN
            WriteFile(hPipe, buf, written, (DWORD*)&written, NULL);
#else
            int snd = ::send(hPipe,buf, written, 0);
            if(snd<0)
            {
                ::close(hPipe);
                hPipe=-1;
            }
#endif
        }
    }

    list.splice(list.begin(),messages[client]);
    return list;
}
